@@ -43,7 +43,7 @@ test_that("Schema$code()", {
     schema(a = int32(), b = struct(c = double(), d = utf8()), e = list_of(binary()))
   )
 
-  skip_if(packageVersion("rlang") < 1)
+  skip_if(packageVersion("rlang") < "1")
   expect_error(
     eval(schema(x = int32(), y = DayTimeInterval__initialize())$code()),
     "Unsupported type"
@@ -276,3 +276,35 @@ test_that("as_schema() works for StructType objects", {
   struct_type <- struct(col1 = int32())
   expect_equal(as_schema(struct_type), schema(col1 = int32()))
 })
+
+test_that("schema name assignment", {
+  schm <- schema(x = int8(), y = string(), z = double())
+  expect_identical(names(schm), c("x", "y", "z"))
+  names(schm) <- c("a", "b", "c")
+  expect_identical(names(schm), c("a", "b", "c"))
+  expect_error(names(schm) <- "f", regexp = "Replacement names must contain same number of items as current names")
+  expect_error(names(schm) <- NULL, regexp = "Replacement names must be character vector, not NULL")
+
+  # Test that R metadata is updated appropriately
+  df <- data.frame(x = 1:3, y = c("a", "b", "c"))
+  schm2 <- arrow_table(df)$schema
+  names(schm2) <- c("col1", "col2")
+  expect_identical(names(schm2), c("col1", "col2"))
+  expect_identical(names(schm2$r_metadata$columns), c("col1", "col2"))
+})
+
+test_that("schema extraction", {
+  skip_if_not_available("dataset")
+  tbl <- arrow_table(example_data)
+  expect_equal(schema(tbl), tbl$schema)
+
+  ds <- InMemoryDataset$create(example_data)
+  expect_equal(schema(ds), ds$schema)
+
+  rdr <- RecordBatchReader$create(record_batch(example_data))
+  expect_equal(schema(rdr), rdr$schema)
+
+  adq <- as_adq(example_data)
+  expect_equal(schema(adq), adq$.data$schema)
+
+})
