@@ -362,9 +362,12 @@ bool Expression::IsSatisfiable() const {
   }
 
   if (call->function_name == "and_kleene" || call->function_name == "and") {
-    for (const Expression& arg : call->arguments) {
-      if (!arg.IsSatisfiable()) return false;
-    }
+    return std::all_of(call->arguments.begin(), call->arguments.end(),
+                       [](const Expression& arg) { return arg.IsSatisfiable(); });
+  }
+  if (call->function_name == "or_kleene" || call->function_name == "or") {
+    return std::any_of(call->arguments.begin(), call->arguments.end(),
+                       [](const Expression& arg) { return arg.IsSatisfiable(); });
   }
 
   return true;
