@@ -27,6 +27,7 @@
 #include "arrow/array/array_binary.h"
 #include "arrow/array/array_nested.h"
 #include "arrow/array/array_primitive.h"
+#include "arrow/array/builder_primitive.h"
 #include "arrow/flight/client_middleware.h"
 #include "arrow/flight/server_middleware.h"
 #include "arrow/flight/sql/client.h"
@@ -37,8 +38,12 @@
 #include "arrow/flight/types.h"
 #include "arrow/ipc/dictionary.h"
 #include "arrow/status.h"
+#include "arrow/table.h"
+#include "arrow/table_builder.h"
 #include "arrow/testing/gtest_util.h"
 #include "arrow/util/checked_cast.h"
+#include "arrow/util/string.h"
+#include "arrow/util/value_parsing.h"
 
 namespace arrow {
 namespace flight {
@@ -143,10 +148,10 @@ class TestServerMiddleware : public ServerMiddleware {
 
 class TestServerMiddlewareFactory : public ServerMiddlewareFactory {
  public:
-  Status StartCall(const CallInfo& info, const CallHeaders& incoming_headers,
+  Status StartCall(const CallInfo& info, const ServerCallContext& context,
                    std::shared_ptr<ServerMiddleware>* middleware) override {
     const std::pair<CallHeaders::const_iterator, CallHeaders::const_iterator>& iter_pair =
-        incoming_headers.equal_range("x-middleware");
+        context.incoming_headers().equal_range("x-middleware");
     std::string received = "";
     if (iter_pair.first != iter_pair.second) {
       const std::string_view& value = (*iter_pair.first).second;
@@ -209,9 +214,10 @@ class MiddlewareServer : public FlightServerBase {
       std::shared_ptr<Schema> schema = arrow::schema({});
       // Return a fake location - the test doesn't read it
       ARROW_ASSIGN_OR_RAISE(auto location, Location::ForGrpcTcp("localhost", 10010));
-      std::vector<FlightEndpoint> endpoints{FlightEndpoint{{"foo"}, {location}}};
-      ARROW_ASSIGN_OR_RAISE(auto info,
-                            FlightInfo::Make(*schema, descriptor, endpoints, -1, -1));
+      std::vector<FlightEndpoint> endpoints{
+          FlightEndpoint{{"foo"}, {location}, std::nullopt}};
+      ARROW_ASSIGN_OR_RAISE(
+          auto info, FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, false));
       *result = std::make_unique<FlightInfo>(info);
       return Status::OK();
     }
@@ -271,6 +277,475 @@ class MiddlewareScenario : public Scenario {
   std::shared_ptr<TestClientMiddlewareFactory> client_middleware_;
 };
 
+/// \brief The server used for testing FlightInfo.ordered.
+///
+/// If the given command is "ordered", the server sets
+/// FlightInfo.ordered. The client that supports FlightInfo.ordered
+/// must read data from endpoints from front to back. The client that
+/// doesn't support FlightInfo.ordered may read data from endpoints in
+/// random order.
+///
+/// This scenario is passed only when the client supports
+/// FlightInfo.ordered.
+class OrderedServer : public FlightServerBase {
+  Status GetFlightInfo(const ServerCallContext& context,
+                       const FlightDescriptor& descriptor,
+                       std::unique_ptr<FlightInfo>* result) override {
+    const auto ordered = (descriptor.type == FlightDescriptor::DescriptorType::CMD &&
+                          descriptor.cmd == "ordered");
+    auto schema = BuildSchema();
+    std::vector<FlightEndpoint> endpoints;
+    if (ordered) {
+      endpoints.push_back(FlightEndpoint{{"1"}, {}, std::nullopt});
+      endpoints.push_back(FlightEndpoint{{"2"}, {}, std::nullopt});
+      endpoints.push_back(FlightEndpoint{{"3"}, {}, std::nullopt});
+    } else {
+      endpoints.push_back(FlightEndpoint{{"1"}, {}, std::nullopt});
+      endpoints.push_back(FlightEndpoint{{"3"}, {}, std::nullopt});
+      endpoints.push_back(FlightEndpoint{{"2"}, {}, std::nullopt});
+    }
+    ARROW_ASSIGN_OR_RAISE(
+        auto info, FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, ordered));
+    *result = std::make_unique<FlightInfo>(info);
+    return Status::OK();
+  }
+
+  Status DoGet(const ServerCallContext& context, const Ticket& request,
+               std::unique_ptr<FlightDataStream>* stream) override {
+    ARROW_ASSIGN_OR_RAISE(auto builder, RecordBatchBuilder::Make(
+                                            BuildSchema(), arrow::default_memory_pool()));
+    auto number_builder = builder->GetFieldAs<Int32Builder>(0);
+    if (request.ticket == "1") {
+      ARROW_RETURN_NOT_OK(number_builder->Append(1));
+      ARROW_RETURN_NOT_OK(number_builder->Append(2));
+      ARROW_RETURN_NOT_OK(number_builder->Append(3));
+    } else if (request.ticket == "2") {
+      ARROW_RETURN_NOT_OK(number_builder->Append(10));
+      ARROW_RETURN_NOT_OK(number_builder->Append(20));
+      ARROW_RETURN_NOT_OK(number_builder->Append(30));
+    } else if (request.ticket == "3") {
+      ARROW_RETURN_NOT_OK(number_builder->Append(100));
+      ARROW_RETURN_NOT_OK(number_builder->Append(200));
+      ARROW_RETURN_NOT_OK(number_builder->Append(300));
+    } else {
+      return Status::KeyError("Could not find flight: ", request.ticket);
+    }
+    ARROW_ASSIGN_OR_RAISE(auto record_batch, builder->Flush());
+    std::vector<std::shared_ptr<RecordBatch>> record_batches{record_batch};
+    ARROW_ASSIGN_OR_RAISE(auto record_batch_reader,
+                          RecordBatchReader::Make(record_batches));
+    *stream = std::make_unique<RecordBatchStream>(record_batch_reader);
+    return Status::OK();
+  }
+
+ private:
+  std::shared_ptr<Schema> BuildSchema() {
+    return arrow::schema({arrow::field("number", arrow::int32(), false)});
+  }
+};
+
+/// \brief The ordered scenario.
+///
+/// This tests that the server and client get expected header values.
+class OrderedScenario : public Scenario {
+  Status MakeServer(std::unique_ptr<FlightServerBase>* server,
+                    FlightServerOptions* options) override {
+    server->reset(new OrderedServer());
+    return Status::OK();
+  }
+
+  Status MakeClient(FlightClientOptions* options) override { return Status::OK(); }
+
+  Status RunClient(std::unique_ptr<FlightClient> client) override {
+    ARROW_ASSIGN_OR_RAISE(auto info,
+                          client->GetFlightInfo(FlightDescriptor::Command("ordered")));
+    if (!info->ordered()) {
+      return Status::Invalid("Server must return FlightInfo.ordered = true");
+    }
+    std::vector<std::shared_ptr<arrow::Table>> tables;
+    for (const auto& endpoint : info->endpoints()) {
+      if (!endpoint.locations.empty()) {
+        std::stringstream ss;
+        ss << "[";
+        for (const auto& location : endpoint.locations) {
+          if (ss.str().size() != 1) {
+            ss << ", ";
+          }
+          ss << location.ToString();
+        }
+        ss << "]";
+        return Status::Invalid(
+            "Expected to receive empty locations to use the original service: ",
+            ss.str());
+      }
+      ARROW_ASSIGN_OR_RAISE(auto reader, client->DoGet(endpoint.ticket));
+      ARROW_ASSIGN_OR_RAISE(auto table, reader->ToTable());
+      tables.push_back(table);
+    }
+    ARROW_ASSIGN_OR_RAISE(auto table, ConcatenateTables(tables));
+
+    // Build expected table
+    auto schema = arrow::schema({arrow::field("number", arrow::int32(), false)});
+    ARROW_ASSIGN_OR_RAISE(auto builder,
+                          RecordBatchBuilder::Make(schema, arrow::default_memory_pool()));
+    auto number_builder = builder->GetFieldAs<Int32Builder>(0);
+    ARROW_RETURN_NOT_OK(number_builder->Append(1));
+    ARROW_RETURN_NOT_OK(number_builder->Append(2));
+    ARROW_RETURN_NOT_OK(number_builder->Append(3));
+    ARROW_RETURN_NOT_OK(number_builder->Append(10));
+    ARROW_RETURN_NOT_OK(number_builder->Append(20));
+    ARROW_RETURN_NOT_OK(number_builder->Append(30));
+    ARROW_RETURN_NOT_OK(number_builder->Append(100));
+    ARROW_RETURN_NOT_OK(number_builder->Append(200));
+    ARROW_RETURN_NOT_OK(number_builder->Append(300));
+    ARROW_ASSIGN_OR_RAISE(auto expected_record_batch, builder->Flush());
+    std::vector<std::shared_ptr<RecordBatch>> expected_record_batches{
+        expected_record_batch};
+    ARROW_ASSIGN_OR_RAISE(auto expected_table,
+                          Table::FromRecordBatches(expected_record_batches));
+
+    // Check read data
+    if (!table->Equals(*expected_table)) {
+      return Status::Invalid("Read data isn't expected\n", "Expected:\n",
+                             expected_table->ToString(), "Actual:\n", table->ToString());
+    }
+    return Status::OK();
+  }
+};
+
+/// \brief The server used for testing FlightEndpoint.expiration_time.
+///
+/// GetFlightInfo() returns a FlightInfo that has the following
+/// three FlightEndpoints:
+///
+/// 1. No expiration time
+/// 2. 5 seconds expiration time
+/// 3. 6 seconds expiration time
+///
+/// The client can't read data from the first endpoint multiple times
+/// but can read data from the second and third endpoints. The client
+/// can't re-read data from the second endpoint 5 seconds later. The
+/// client can't re-read data from the third endpoint 6 seconds
+/// later.
+///
+/// The client can cancel a returned FlightInfo by pre-defined
+/// CancelFlightInfo action. The client can't read data from endpoints
+/// even within 6 seconds after the action.
+///
+/// The client can extend the expiration time of a FlightEndpoint in
+/// a returned FlightInfo by pre-defined RenewFlightEndpoint
+/// action. The client can read data from endpoints multiple times
+/// within more 10 seconds after the action.
+class ExpirationTimeServer : public FlightServerBase {
+ private:
+  struct EndpointStatus {
+    explicit EndpointStatus(std::optional<Timestamp> expiration_time)
+        : expiration_time(expiration_time) {}
+
+    std::optional<Timestamp> expiration_time;
+    uint32_t num_gets = 0;
+    bool cancelled = false;
+  };
+
+ public:
+  ExpirationTimeServer() : FlightServerBase(), statuses_() {}
+
+  Status GetFlightInfo(const ServerCallContext& context,
+                       const FlightDescriptor& descriptor,
+                       std::unique_ptr<FlightInfo>* result) override {
+    statuses_.clear();
+    auto schema = BuildSchema();
+    std::vector<FlightEndpoint> endpoints;
+    AddEndpoint(endpoints, "No expiration time", std::nullopt);
+    AddEndpoint(endpoints, "5 seconds",
+                Timestamp::clock::now() + std::chrono::seconds{5});
+    AddEndpoint(endpoints, "6 seconds",
+                Timestamp::clock::now() + std::chrono::seconds{6});
+    ARROW_ASSIGN_OR_RAISE(
+        auto info, FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, false));
+    *result = std::make_unique<FlightInfo>(info);
+    return Status::OK();
+  }
+
+  Status DoGet(const ServerCallContext& context, const Ticket& request,
+               std::unique_ptr<FlightDataStream>* stream) override {
+    ARROW_ASSIGN_OR_RAISE(auto index, ExtractIndexFromTicket(request.ticket));
+    auto& status = statuses_[index];
+    if (status.cancelled) {
+      return Status::KeyError("Invalid flight: canceled: ", request.ticket);
+    }
+    if (status.expiration_time.has_value()) {
+      auto expiration_time = status.expiration_time.value();
+      if (expiration_time < Timestamp::clock::now()) {
+        return Status::KeyError("Invalid flight: expired: ", request.ticket);
+      }
+    } else {
+      if (status.num_gets > 0) {
+        return Status::KeyError("Invalid flight: can't read multiple times: ",
+                                request.ticket);
+      }
+    }
+    status.num_gets++;
+    ARROW_ASSIGN_OR_RAISE(auto builder, RecordBatchBuilder::Make(
+                                            BuildSchema(), arrow::default_memory_pool()));
+    auto number_builder = builder->GetFieldAs<UInt32Builder>(0);
+    ARROW_RETURN_NOT_OK(number_builder->Append(index));
+    ARROW_ASSIGN_OR_RAISE(auto record_batch, builder->Flush());
+    std::vector<std::shared_ptr<RecordBatch>> record_batches{record_batch};
+    ARROW_ASSIGN_OR_RAISE(auto record_batch_reader,
+                          RecordBatchReader::Make(record_batches));
+    *stream = std::make_unique<RecordBatchStream>(record_batch_reader);
+    return Status::OK();
+  }
+
+  Status DoAction(const ServerCallContext& context, const Action& action,
+                  std::unique_ptr<ResultStream>* result_stream) override {
+    std::vector<Result> results;
+    if (action.type == ActionType::kCancelFlightInfo.type) {
+      ARROW_ASSIGN_OR_RAISE(auto request, CancelFlightInfoRequest::Deserialize(
+                                              std::string_view(*action.body)));
+      auto cancel_status = CancelStatus::kUnspecified;
+      for (const auto& endpoint : request.info->endpoints()) {
+        auto index_result = ExtractIndexFromTicket(endpoint.ticket.ticket);
+        if (index_result.ok()) {
+          auto index = *index_result;
+          if (statuses_[index].cancelled) {
+            cancel_status = CancelStatus::kNotCancellable;
+          } else {
+            statuses_[index].cancelled = true;
+            if (cancel_status == CancelStatus::kUnspecified) {
+              cancel_status = CancelStatus::kCancelled;
+            }
+          }
+        } else {
+          cancel_status = CancelStatus::kNotCancellable;
+        }
+      }
+      CancelFlightInfoResult cancel_result{cancel_status};
+      ARROW_ASSIGN_OR_RAISE(auto serialized, cancel_result.SerializeToString());
+      results.push_back(Result{Buffer::FromString(std::move(serialized))});
+    } else if (action.type == ActionType::kRenewFlightEndpoint.type) {
+      ARROW_ASSIGN_OR_RAISE(auto request, RenewFlightEndpointRequest::Deserialize(
+                                              std::string_view(*action.body)));
+      auto& endpoint = request.endpoint;
+      ARROW_ASSIGN_OR_RAISE(auto index, ExtractIndexFromTicket(endpoint.ticket.ticket));
+      if (statuses_[index].cancelled) {
+        return Status::Invalid("Invalid flight: canceled: ", endpoint.ticket.ticket);
+      }
+      endpoint.ticket.ticket += ": renewed (+ 10 seconds)";
+      endpoint.expiration_time = Timestamp::clock::now() + std::chrono::seconds{10};
+      statuses_[index].expiration_time = endpoint.expiration_time.value();
+      ARROW_ASSIGN_OR_RAISE(auto serialized, endpoint.SerializeToString());
+      results.push_back(Result{Buffer::FromString(std::move(serialized))});
+    } else {
+      return Status::Invalid("Unknown action: ", action.type);
+    }
+    *result_stream = std::make_unique<SimpleResultStream>(std::move(results));
+    return Status::OK();
+  }
+
+  Status ListActions(const ServerCallContext& context,
+                     std::vector<ActionType>* actions) override {
+    *actions = {
+        ActionType::kCancelFlightInfo,
+        ActionType::kRenewFlightEndpoint,
+    };
+    return Status::OK();
+  }
+
+ private:
+  void AddEndpoint(std::vector<FlightEndpoint>& endpoints, std::string ticket,
+                   std::optional<Timestamp> expiration_time) {
+    endpoints.push_back(FlightEndpoint{
+        {std::to_string(statuses_.size()) + ": " + ticket}, {}, expiration_time});
+    statuses_.emplace_back(expiration_time);
+  }
+
+  arrow::Result<uint32_t> ExtractIndexFromTicket(const std::string& ticket) {
+    auto index_string = arrow::internal::SplitString(ticket, ':', 2)[0];
+    uint32_t index;
+    if (!arrow::internal::ParseUnsigned(index_string.data(), index_string.length(),
+                                        &index)) {
+      return Status::KeyError("Invalid flight: no index: ", ticket);
+    }
+    if (index >= statuses_.size()) {
+      return Status::KeyError("Invalid flight: out of index: ", ticket);
+    }
+    return index;
+  }
+
+  std::shared_ptr<Schema> BuildSchema() {
+    return arrow::schema({arrow::field("number", arrow::uint32(), false)});
+  }
+
+  std::vector<EndpointStatus> statuses_;
+};
+
+/// \brief The expiration time scenario - DoGet.
+///
+/// This tests that the client can read data that isn't expired yet
+/// multiple times and can't read data after it's expired.
+class ExpirationTimeDoGetScenario : public Scenario {
+  Status MakeServer(std::unique_ptr<FlightServerBase>* server,
+                    FlightServerOptions* options) override {
+    *server = std::make_unique<ExpirationTimeServer>();
+    return Status::OK();
+  }
+
+  Status MakeClient(FlightClientOptions* options) override { return Status::OK(); }
+
+  Status RunClient(std::unique_ptr<FlightClient> client) override {
+    ARROW_ASSIGN_OR_RAISE(
+        auto info, client->GetFlightInfo(FlightDescriptor::Command("expiration_time")));
+    std::vector<std::shared_ptr<arrow::Table>> tables;
+    for (const auto& endpoint : info->endpoints()) {
+      if (tables.size() == 0) {
+        if (endpoint.expiration_time.has_value()) {
+          return Status::Invalid("endpoints[0] must not have expiration time");
+        }
+      } else {
+        if (!endpoint.expiration_time.has_value()) {
+          return Status::Invalid("endpoints[", tables.size(),
+                                 "] must have expiration time");
+        }
+      }
+      ARROW_ASSIGN_OR_RAISE(auto reader, client->DoGet(endpoint.ticket));
+      ARROW_ASSIGN_OR_RAISE(auto table, reader->ToTable());
+      tables.push_back(table);
+    }
+    ARROW_ASSIGN_OR_RAISE(auto table, ConcatenateTables(tables));
+
+    // Build expected table
+    auto schema = arrow::schema({arrow::field("number", arrow::uint32(), false)});
+    ARROW_ASSIGN_OR_RAISE(auto builder,
+                          RecordBatchBuilder::Make(schema, arrow::default_memory_pool()));
+    auto number_builder = builder->GetFieldAs<UInt32Builder>(0);
+    ARROW_RETURN_NOT_OK(number_builder->Append(0));
+    ARROW_RETURN_NOT_OK(number_builder->Append(1));
+    ARROW_RETURN_NOT_OK(number_builder->Append(2));
+    ARROW_ASSIGN_OR_RAISE(auto expected_record_batch, builder->Flush());
+    std::vector<std::shared_ptr<RecordBatch>> expected_record_batches{
+        expected_record_batch};
+    ARROW_ASSIGN_OR_RAISE(auto expected_table,
+                          Table::FromRecordBatches(expected_record_batches));
+
+    // Check read data
+    if (!table->Equals(*expected_table)) {
+      return Status::Invalid("Read data isn't expected\n", "Expected:\n",
+                             expected_table->ToString(), "Actual:\n", table->ToString());
+    }
+    return Status::OK();
+  }
+};
+
+/// \brief The expiration time scenario - ListActions.
+///
+/// This tests that the client can get pre-defined actions and the
+/// server uses pre-defined ActionTypes for ListActions.
+class ExpirationTimeListActionsScenario : public Scenario {
+  Status MakeServer(std::unique_ptr<FlightServerBase>* server,
+                    FlightServerOptions* options) override {
+    *server = std::make_unique<ExpirationTimeServer>();
+    return Status::OK();
+  }
+
+  Status MakeClient(FlightClientOptions* options) override { return Status::OK(); }
+
+  Status RunClient(std::unique_ptr<FlightClient> client) override {
+    ARROW_ASSIGN_OR_RAISE(auto action_types, client->ListActions());
+    std::vector<std::string> actual_action_types;
+    for (const auto& action_type : action_types) {
+      actual_action_types.push_back(action_type.type);
+    }
+    std::sort(actual_action_types.begin(), actual_action_types.end());
+    std::vector<std::string> expected_action_types = {
+        "CancelFlightInfo",
+        "RenewFlightEndpoint",
+    };
+    if (actual_action_types != expected_action_types) {
+      return Status::Invalid(
+          "Invalid ListActions response: expected=[",
+          arrow::internal::JoinStrings(expected_action_types, ", "), "] actual=[",
+          arrow::internal::JoinStrings(actual_action_types, ", "), "]");
+    }
+    return Status::OK();
+  }
+};
+
+/// \brief The expiration time scenario - CancelFlightInfo.
+///
+/// This tests that the client can cancel a FlightInfo explicitly and
+/// the server returns an error for DoGet against endpoints in the
+/// cancelled FlightInfo.
+class ExpirationTimeCancelFlightInfoScenario : public Scenario {
+  Status MakeServer(std::unique_ptr<FlightServerBase>* server,
+                    FlightServerOptions* options) override {
+    *server = std::make_unique<ExpirationTimeServer>();
+    return Status::OK();
+  }
+
+  Status MakeClient(FlightClientOptions* options) override { return Status::OK(); }
+
+  Status RunClient(std::unique_ptr<FlightClient> client) override {
+    ARROW_ASSIGN_OR_RAISE(auto info,
+                          client->GetFlightInfo(FlightDescriptor::Command("expiration")));
+    CancelFlightInfoRequest request{std::move(info)};
+    ARROW_ASSIGN_OR_RAISE(auto cancel_result, client->CancelFlightInfo(request));
+    if (cancel_result.status != CancelStatus::kCancelled) {
+      return Status::Invalid("CancelFlightInfo must return CANCEL_STATUS_CANCELLED: ",
+                             cancel_result.ToString());
+    }
+    info = std::move(request.info);
+    for (const auto& endpoint : info->endpoints()) {
+      auto reader = client->DoGet(endpoint.ticket);
+      if (reader.ok()) {
+        return Status::Invalid("DoGet after CancelFlightInfo must be failed");
+      }
+    }
+    return Status::OK();
+  }
+};
+
+/// \brief The expiration time scenario - RenewFlightEndpoint.
+///
+/// This tests that the client can renew a FlightEndpoint and read
+/// data in renewed expiration time even when the original
+/// expiration time is over.
+class ExpirationTimeRenewFlightEndpointScenario : public Scenario {
+  Status MakeServer(std::unique_ptr<FlightServerBase>* server,
+                    FlightServerOptions* options) override {
+    *server = std::make_unique<ExpirationTimeServer>();
+    return Status::OK();
+  }
+
+  Status MakeClient(FlightClientOptions* options) override { return Status::OK(); }
+
+  Status RunClient(std::unique_ptr<FlightClient> client) override {
+    ARROW_ASSIGN_OR_RAISE(auto info,
+                          client->GetFlightInfo(FlightDescriptor::Command("expiration")));
+    // Renew all endpoints that have expiration time
+    for (const auto& endpoint : info->endpoints()) {
+      if (!endpoint.expiration_time.has_value()) {
+        continue;
+      }
+      const auto& expiration_time = endpoint.expiration_time.value();
+      auto request = RenewFlightEndpointRequest{endpoint};
+      ARROW_ASSIGN_OR_RAISE(auto renewed_endpoint, client->RenewFlightEndpoint(request));
+      if (!renewed_endpoint.expiration_time.has_value()) {
+        return Status::Invalid("Renewed endpoint must have expiration time: ",
+                               renewed_endpoint.ToString());
+      }
+      const auto& renewed_expiration_time = renewed_endpoint.expiration_time.value();
+      if (renewed_expiration_time <= expiration_time) {
+        return Status::Invalid("Renewed endpoint must have newer expiration time\n",
+                               "Original:\n", endpoint.ToString(), "Renewed:\n",
+                               renewed_endpoint.ToString());
+      }
+    }
+    return Status::OK();
+  }
+};
+
 /// \brief Schema to be returned for mocking the statement/prepared statement results.
 ///
 /// Must be the same across all languages.
@@ -381,9 +856,9 @@ class FlightSqlScenarioServer : public sql::FlightSqlServerBase {
       schema = GetQueryWithTransactionSchema().get();
     }
     ARROW_ASSIGN_OR_RAISE(auto handle, sql::CreateStatementQueryTicket(ticket));
-    std::vector<FlightEndpoint> endpoints{FlightEndpoint{{handle}, {}}};
-    ARROW_ASSIGN_OR_RAISE(auto result,
-                          FlightInfo::Make(*schema, descriptor, endpoints, -1, -1));
+    std::vector<FlightEndpoint> endpoints{FlightEndpoint{{handle}, {}, std::nullopt}};
+    ARROW_ASSIGN_OR_RAISE(
+        auto result, FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, false));
     return std::make_unique<FlightInfo>(result);
   }
 
@@ -406,9 +881,9 @@ class FlightSqlScenarioServer : public sql::FlightSqlServerBase {
       schema = GetQueryWithTransactionSchema().get();
     }
     ARROW_ASSIGN_OR_RAISE(auto handle, sql::CreateStatementQueryTicket(ticket));
-    std::vector<FlightEndpoint> endpoints{FlightEndpoint{{handle}, {}}};
-    ARROW_ASSIGN_OR_RAISE(auto result,
-                          FlightInfo::Make(*schema, descriptor, endpoints, -1, -1));
+    std::vector<FlightEndpoint> endpoints{FlightEndpoint{{handle}, {}, std::nullopt}};
+    ARROW_ASSIGN_OR_RAISE(
+        auto result, FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, false));
     return std::make_unique<FlightInfo>(result);
   }
 
@@ -803,17 +1278,17 @@ class FlightSqlScenarioServer : public sql::FlightSqlServerBase {
     return sql::ActionBeginTransactionResult{kTransactionId};
   }
 
-  arrow::Result<sql::CancelResult> CancelQuery(
-      const ServerCallContext& context,
-      const sql::ActionCancelQueryRequest& request) override {
-    ARROW_RETURN_NOT_OK(AssertEq<size_t>(1, request.info->endpoints().size(),
-                                         "Expected 1 endpoint for CancelQuery"));
-    const FlightEndpoint& endpoint = request.info->endpoints()[0];
+  arrow::Result<CancelFlightInfoResult> CancelFlightInfo(
+      const ServerCallContext& context, const CancelFlightInfoRequest& request) override {
+    const auto& info = request.info;
+    ARROW_RETURN_NOT_OK(AssertEq<size_t>(1, info->endpoints().size(),
+                                         "Expected 1 endpoint for CancelFlightInfo"));
+    const auto& endpoint = info->endpoints()[0];
     ARROW_ASSIGN_OR_RAISE(auto ticket,
                           sql::StatementQueryTicket::Deserialize(endpoint.ticket.ticket));
     ARROW_RETURN_NOT_OK(AssertEq<std::string>("PLAN HANDLE", ticket.statement_handle,
-                                              "Unexpected ticket in CancelQuery"));
-    return sql::CancelResult::kCancelled;
+                                              "Unexpected ticket in CancelFlightInfo"));
+    return CancelFlightInfoResult{CancelStatus::kCancelled};
   }
 
   Status EndSavepoint(const ServerCallContext& context,
@@ -849,9 +1324,10 @@ class FlightSqlScenarioServer : public sql::FlightSqlServerBase {
  private:
   arrow::Result<std::unique_ptr<FlightInfo>> GetFlightInfoForCommand(
       const FlightDescriptor& descriptor, const std::shared_ptr<Schema>& schema) {
-    std::vector<FlightEndpoint> endpoints{FlightEndpoint{{descriptor.cmd}, {}}};
+    std::vector<FlightEndpoint> endpoints{
+        FlightEndpoint{{descriptor.cmd}, {}, std::nullopt}};
     ARROW_ASSIGN_OR_RAISE(auto result,
-                          FlightInfo::Make(*schema, descriptor, endpoints, -1, -1))
+                          FlightInfo::Make(*schema, descriptor, endpoints, -1, -1, false))
 
     return std::make_unique<FlightInfo>(result);
   }
@@ -1177,8 +1653,12 @@ class FlightSqlExtensionScenario : public FlightSqlScenario {
     ARROW_RETURN_NOT_OK(ValidateSchema(GetQuerySchema(), *schema));
 
     ARROW_ASSIGN_OR_RAISE(info, sql_client->ExecuteSubstrait({}, kSubstraitPlan));
-    ARROW_ASSIGN_OR_RAISE(sql::CancelResult cancel_result,
-                          sql_client->CancelQuery({}, *info));
+    // TODO: Use CancelFLightInfo() instead of CancelQuery() here. We
+    // use CancelQuery() here for now because some Flight SQL
+    // implementations still don't support CancelFlightInfo yet.
+    ARROW_SUPPRESS_DEPRECATION_WARNING
+    ARROW_ASSIGN_OR_RAISE(auto cancel_result, sql_client->CancelQuery({}, *info));
+    ARROW_UNSUPPRESS_DEPRECATION_WARNING
     ARROW_RETURN_NOT_OK(
         AssertEq(sql::CancelResult::kCancelled, cancel_result, "Wrong cancel result"));
 
@@ -1330,6 +1810,21 @@ Status GetScenario(const std::string& scenario_name, std::shared_ptr<Scenario>*
   } else if (scenario_name == "middleware") {
     *out = std::make_shared<MiddlewareScenario>();
     return Status::OK();
+  } else if (scenario_name == "ordered") {
+    *out = std::make_shared<OrderedScenario>();
+    return Status::OK();
+  } else if (scenario_name == "expiration_time:do_get") {
+    *out = std::make_shared<ExpirationTimeDoGetScenario>();
+    return Status::OK();
+  } else if (scenario_name == "expiration_time:list_actions") {
+    *out = std::make_shared<ExpirationTimeListActionsScenario>();
+    return Status::OK();
+  } else if (scenario_name == "expiration_time:cancel_flight_info") {
+    *out = std::make_shared<ExpirationTimeCancelFlightInfoScenario>();
+    return Status::OK();
+  } else if (scenario_name == "expiration_time:renew_flight_endpoint") {
+    *out = std::make_shared<ExpirationTimeRenewFlightEndpointScenario>();
+    return Status::OK();
   } else if (scenario_name == "flight_sql") {
     *out = std::make_shared<FlightSqlScenario>();
     return Status::OK();
