@@ -27,6 +27,7 @@
 
 #include "arrow/buffer.h"
 #include "arrow/builder.h"
+#include "arrow/flight/serialization_internal.h"
 #include "arrow/flight/sql/protocol_internal.h"
 #include "arrow/flight/sql/sql_info_internal.h"
 #include "arrow/type.h"
@@ -361,14 +362,22 @@ arrow::Result<ActionEndTransactionRequest> ParseActionEndTransactionRequest(
 
 arrow::Result<Result> PackActionResult(const google::protobuf::Message& message) {
   google::protobuf::Any any;
+#if PROTOBUF_VERSION >= 3015000
   if (!any.PackFrom(message)) {
     return Status::IOError("Failed to pack ", message.GetTypeName());
   }
+#else
+  any.PackFrom(message);
+#endif
 
   std::string buffer;
+#if PROTOBUF_VERSION >= 3015000
   if (!any.SerializeToString(&buffer)) {
     return Status::IOError("Failed to serialize packed ", message.GetTypeName());
   }
+#else
+  any.SerializeToString(&buffer);
+#endif
   return Result{Buffer::FromString(std::move(buffer))};
 }
 
@@ -384,6 +393,16 @@ arrow::Result<Result> PackActionResult(ActionBeginTransactionResult result) {
   return PackActionResult(pb_result);
 }
 
+arrow::Result<Result> PackActionResult(CancelFlightInfoResult result) {
+  ARROW_ASSIGN_OR_RAISE(auto serialized, result.SerializeToString());
+  return Result{Buffer::FromString(std::move(serialized))};
+}
+
+arrow::Result<Result> PackActionResult(const FlightEndpoint& endpoint) {
+  ARROW_ASSIGN_OR_RAISE(auto serialized, endpoint.SerializeToString());
+  return Result{Buffer::FromString(std::move(serialized))};
+}
+
 arrow::Result<Result> PackActionResult(CancelResult result) {
   pb::sql::ActionCancelQueryResult pb_result;
   switch (result) {
@@ -741,6 +760,8 @@ Status FlightSqlServerBase::DoPut(const ServerCallContext& context,
 Status FlightSqlServerBase::ListActions(const ServerCallContext& context,
                                         std::vector<ActionType>* actions) {
   *actions = {
+      ActionType::kCancelFlightInfo,
+      ActionType::kRenewFlightEndpoint,
       FlightSqlServerBase::kBeginSavepointActionType,
       FlightSqlServerBase::kBeginTransactionActionType,
       FlightSqlServerBase::kCancelQueryActionType,
@@ -756,67 +777,86 @@ Status FlightSqlServerBase::ListActions(const ServerCallContext& context,
 Status FlightSqlServerBase::DoAction(const ServerCallContext& context,
                                      const Action& action,
                                      std::unique_ptr<ResultStream>* result_stream) {
-  google::protobuf::Any any;
-  if (!any.ParseFromArray(action.body->data(), static_cast<int>(action.body->size()))) {
-    return Status::Invalid("Unable to parse action");
-  }
-
   std::vector<Result> results;
-  if (action.type == FlightSqlServerBase::kBeginSavepointActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionBeginSavepointRequest internal_command,
-                          ParseActionBeginSavepointRequest(any));
-    ARROW_ASSIGN_OR_RAISE(ActionBeginSavepointResult result,
-                          BeginSavepoint(context, internal_command));
-    ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+  if (action.type == ActionType::kCancelFlightInfo.type) {
+    std::string_view body(*action.body);
+    ARROW_ASSIGN_OR_RAISE(auto request, CancelFlightInfoRequest::Deserialize(body));
+    ARROW_ASSIGN_OR_RAISE(auto result, CancelFlightInfo(context, request));
+    ARROW_ASSIGN_OR_RAISE(auto packed_result, PackActionResult(std::move(result)));
 
     results.push_back(std::move(packed_result));
-  } else if (action.type == FlightSqlServerBase::kBeginTransactionActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionBeginTransactionRequest internal_command,
-                          ParseActionBeginTransactionRequest(any));
-    ARROW_ASSIGN_OR_RAISE(ActionBeginTransactionResult result,
-                          BeginTransaction(context, internal_command));
-    ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+  } else if (action.type == ActionType::kRenewFlightEndpoint.type) {
+    std::string_view body(*action.body);
+    ARROW_ASSIGN_OR_RAISE(auto request, RenewFlightEndpointRequest::Deserialize(body));
+    ARROW_ASSIGN_OR_RAISE(auto renewed_endpoint, RenewFlightEndpoint(context, request));
+    ARROW_ASSIGN_OR_RAISE(auto packed_result, PackActionResult(renewed_endpoint));
 
     results.push_back(std::move(packed_result));
-  } else if (action.type == FlightSqlServerBase::kCancelQueryActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionCancelQueryRequest internal_command,
-                          ParseActionCancelQueryRequest(any));
-    ARROW_ASSIGN_OR_RAISE(CancelResult result, CancelQuery(context, internal_command));
-    ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(result));
-
-    results.push_back(std::move(packed_result));
-  } else if (action.type ==
-             FlightSqlServerBase::kCreatePreparedStatementActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementRequest internal_command,
-                          ParseActionCreatePreparedStatementRequest(any));
-    ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementResult result,
-                          CreatePreparedStatement(context, internal_command));
-    ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
-
-    results.push_back(std::move(packed_result));
-  } else if (action.type ==
-             FlightSqlServerBase::kCreatePreparedSubstraitPlanActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedSubstraitPlanRequest internal_command,
-                          ParseActionCreatePreparedSubstraitPlanRequest(any));
-    ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementResult result,
-                          CreatePreparedSubstraitPlan(context, internal_command));
-    ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
-
-    results.push_back(std::move(packed_result));
-  } else if (action.type == FlightSqlServerBase::kClosePreparedStatementActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionClosePreparedStatementRequest internal_command,
-                          ParseActionClosePreparedStatementRequest(any));
-    ARROW_RETURN_NOT_OK(ClosePreparedStatement(context, internal_command));
-  } else if (action.type == FlightSqlServerBase::kEndSavepointActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionEndSavepointRequest internal_command,
-                          ParseActionEndSavepointRequest(any));
-    ARROW_RETURN_NOT_OK(EndSavepoint(context, internal_command));
-  } else if (action.type == FlightSqlServerBase::kEndTransactionActionType.type) {
-    ARROW_ASSIGN_OR_RAISE(ActionEndTransactionRequest internal_command,
-                          ParseActionEndTransactionRequest(any));
-    ARROW_RETURN_NOT_OK(EndTransaction(context, internal_command));
   } else {
-    return Status::NotImplemented("Action not implemented: ", action.type);
+    google::protobuf::Any any;
+    if (!any.ParseFromArray(action.body->data(), static_cast<int>(action.body->size()))) {
+      return Status::Invalid("Unable to parse action");
+    }
+
+    if (action.type == FlightSqlServerBase::kBeginSavepointActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionBeginSavepointRequest internal_command,
+                            ParseActionBeginSavepointRequest(any));
+      ARROW_ASSIGN_OR_RAISE(ActionBeginSavepointResult result,
+                            BeginSavepoint(context, internal_command));
+      ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+
+      results.push_back(std::move(packed_result));
+    } else if (action.type == FlightSqlServerBase::kBeginTransactionActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionBeginTransactionRequest internal_command,
+                            ParseActionBeginTransactionRequest(any));
+      ARROW_ASSIGN_OR_RAISE(ActionBeginTransactionResult result,
+                            BeginTransaction(context, internal_command));
+      ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+
+      results.push_back(std::move(packed_result));
+    } else if (action.type == FlightSqlServerBase::kCancelQueryActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionCancelQueryRequest internal_command,
+                            ParseActionCancelQueryRequest(any));
+      ARROW_SUPPRESS_DEPRECATION_WARNING
+      ARROW_ASSIGN_OR_RAISE(CancelResult result, CancelQuery(context, internal_command));
+      ARROW_UNSUPPRESS_DEPRECATION_WARNING
+      ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(result));
+
+      results.push_back(std::move(packed_result));
+    } else if (action.type ==
+               FlightSqlServerBase::kCreatePreparedStatementActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementRequest internal_command,
+                            ParseActionCreatePreparedStatementRequest(any));
+      ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementResult result,
+                            CreatePreparedStatement(context, internal_command));
+      ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+
+      results.push_back(std::move(packed_result));
+    } else if (action.type ==
+               FlightSqlServerBase::kCreatePreparedSubstraitPlanActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedSubstraitPlanRequest internal_command,
+                            ParseActionCreatePreparedSubstraitPlanRequest(any));
+      ARROW_ASSIGN_OR_RAISE(ActionCreatePreparedStatementResult result,
+                            CreatePreparedSubstraitPlan(context, internal_command));
+      ARROW_ASSIGN_OR_RAISE(Result packed_result, PackActionResult(std::move(result)));
+
+      results.push_back(std::move(packed_result));
+    } else if (action.type ==
+               FlightSqlServerBase::kClosePreparedStatementActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionClosePreparedStatementRequest internal_command,
+                            ParseActionClosePreparedStatementRequest(any));
+      ARROW_RETURN_NOT_OK(ClosePreparedStatement(context, internal_command));
+    } else if (action.type == FlightSqlServerBase::kEndSavepointActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionEndSavepointRequest internal_command,
+                            ParseActionEndSavepointRequest(any));
+      ARROW_RETURN_NOT_OK(EndSavepoint(context, internal_command));
+    } else if (action.type == FlightSqlServerBase::kEndTransactionActionType.type) {
+      ARROW_ASSIGN_OR_RAISE(ActionEndTransactionRequest internal_command,
+                            ParseActionEndTransactionRequest(any));
+      ARROW_RETURN_NOT_OK(EndTransaction(context, internal_command));
+    } else {
+      return Status::NotImplemented("Action not implemented: ", action.type);
+    }
   }
   *result_stream = std::make_unique<SimpleResultStream>(std::move(results));
   return Status::OK();
@@ -889,9 +929,11 @@ arrow::Result<std::unique_ptr<FlightInfo>> FlightSqlServerBase::GetFlightInfoSql
     return Status::KeyError("No SQL information available.");
   }
 
-  std::vector<FlightEndpoint> endpoints{FlightEndpoint{{descriptor.cmd}, {}}};
-  ARROW_ASSIGN_OR_RAISE(auto result, FlightInfo::Make(*SqlSchema::GetSqlInfoSchema(),
-                                                      descriptor, endpoints, -1, -1))
+  std::vector<FlightEndpoint> endpoints{
+      FlightEndpoint{{descriptor.cmd}, {}, std::nullopt}};
+  ARROW_ASSIGN_OR_RAISE(
+      auto result, FlightInfo::Make(*SqlSchema::GetSqlInfoSchema(), descriptor, endpoints,
+                                    -1, -1, false))
 
   return std::make_unique<FlightInfo>(result);
 }
@@ -1037,9 +1079,23 @@ arrow::Result<ActionBeginTransactionResult> FlightSqlServerBase::BeginTransactio
   return Status::NotImplemented("BeginTransaction not implemented");
 }
 
+arrow::Result<CancelFlightInfoResult> FlightSqlServerBase::CancelFlightInfo(
+    const ServerCallContext& context, const CancelFlightInfoRequest& request) {
+  return Status::NotImplemented("CancelFlightInfo not implemented");
+}
+
 arrow::Result<CancelResult> FlightSqlServerBase::CancelQuery(
     const ServerCallContext& context, const ActionCancelQueryRequest& request) {
-  return Status::NotImplemented("CancelQuery not implemented");
+  CancelFlightInfoRequest cancel_flight_info_request;
+  cancel_flight_info_request.info = std::make_unique<FlightInfo>(*request.info);
+  ARROW_ASSIGN_OR_RAISE(auto result,
+                        CancelFlightInfo(context, cancel_flight_info_request));
+  return static_cast<CancelResult>(result.status);
+}
+
+arrow::Result<FlightEndpoint> FlightSqlServerBase::RenewFlightEndpoint(
+    const ServerCallContext& context, const RenewFlightEndpointRequest& request) {
+  return Status::NotImplemented("RenewFlightEndpoint not implemented");
 }
 
 arrow::Result<ActionCreatePreparedStatementResult>
